# -*- coding: utf-8 -*-
# Copyright 2019, The Emissions API Developers
# https://emissions-api.org
# This software is available under the terms of an MIT license.
# See LICENSE fore more information.
'''Sentinel-5P Downloader
'''

import hashlib
import io
import json
import os.path
import pycurl
import urllib.parse
import logging
import time

# Data publicly provided by ESA:
API = 'https://s5phub.copernicus.eu/dhus/'
USER = 's5pguest'
PASS = 's5pguest'

logger = logging.getLogger(__name__)
'''Logger used by ``sentinel5dl``. Use this to specifically modify the
libraries log level like this::

    sentinel5dl.logger.setLevel(logging.DEBUG)
'''

ca_info = None
'''Path to Certificate Authority (CA) bundle. If this is set,
the value is passed to CURLOPT_CAINFO. If not set,
this option is by default set to the system path
where libcurl's cacert bundle is assumed to be stored,
as established at build time. If this is not already supplied
by your operating system, certifi provides an easy way of
providing a cabundle.
'''


def __md5(filename):
    '''Generate the md5 sum of a file

    :param filename: input filename for which the md5 sum is generated.
    :type filename: str
    :returns: hex representation of the md5 sum with uppercase characters.
    :rtype: str
    '''
    hash_md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash_md5.update(chunk)
    return hash_md5.hexdigest().upper()


def __check_md5(filename, base_path):
    '''Check the md5 sum of a given file against the ESA API.

    :param filename: Path of local file to check
    :param base_path: Base API path to for this product
    :returns: If the local file matches the md5 checksum
    :rtype: bool
    '''
    md5file = f'{filename}.md5sum'
    try:
        with open(md5file, 'r') as f:
            md5sum = f.read()
    except FileNotFoundError:
        md5sum = __http_request(f'{base_path}/Checksum/Value/$value')
        md5sum = md5sum.decode('ascii')
        with open(md5file, 'w') as f:
            f.write(md5sum)

    # Compare md5 sum
    return __md5(filename) == md5sum


def __http_request(path, filename=None, retries=9):
    '''Make an HTTP request to the API via HTTP, optionally downloading the
    response.

    :param path: Request path relative to the base API.
    :param filename: Optional output file name. Note that this file will be
                     overwritten if it already exists. If no filename is
                     provided, the response will be returned.
    :param retries: Number of times the request should be repeated if an error
                    occurred with that request (e.g. a network timeout)
    :returns: The response body or None if a filename is provided.
    '''

    url = API + path.lstrip('/')
    logger.debug('Requesting %s', url)
    try:
        with open(f'{filename}.tmp', 'wb') if filename else io.BytesIO() as f:
            curl = pycurl.Curl()
            curl.setopt(curl.URL, url.encode('ascii', 'ignore'))
            curl.setopt(curl.USERPWD, f'{USER}:{PASS}')
            curl.setopt(curl.WRITEDATA, f)
            curl.setopt(curl.FAILONERROR, True)

            # Use a Certificate Authority (CA) bundle if set
            if ca_info:
                curl.setopt(pycurl.CAINFO, ca_info)

            # Abort if data transfer is not responding but didn't errored
            curl.setopt(pycurl.LOW_SPEED_TIME, 60)
            curl.setopt(pycurl.LOW_SPEED_LIMIT, 30)

            curl.perform()
            curl.close()

            if filename:
                os.rename(f'{filename}.tmp', filename)
            else:
                return f.getvalue()

    except pycurl.error as err:
        if not retries:
            if filename:
                logger.info('Removing temporary file %s.tmp', filename)
                os.remove(f'{filename}.tmp')
            raise err
        logger.warning('Retrying failed HTTP request. %s', err)
        time.sleep(1)
        __http_request(path, filename, retries-1)


def _search(polygon, begin_ts, end_ts, product, processing_level,
            processing_mode, offset, limit):
    '''Make a single search request for products to the API.

    :param polygon: WKT polygon specifying an area the data should intersect
    :param begin_ts: ISO-8601 timestamp specifying the earliest sensing date
    :param end_ts: ISO-8601 timestamp specifying the latest sensing date
    :param product: Type of product to request
    :param processing_level: Data processing level (``L1B`` or ``L2``)
    :param processing_mode: Data processing mode (``Offline``,
                            ``Near real time`` or ``Reprocessing``)
    :param offset: Offset for the results to return
    :param limit: Limit number of results
    :returns: Dictionary containing information about found products
    '''
    filter_query = ['platformname:Sentinel-5']
    if polygon:
        filter_query.append(f'footprint:"Intersects({polygon})"')
    if begin_ts:
        filter_query.append(f'beginPosition:[{begin_ts} TO {end_ts}]')
    if end_ts:
        filter_query.append(f'endPosition:[{begin_ts} TO {end_ts}]',)
    if product:
        filter_query.append(f'producttype:{product}')
    if processing_level:
        filter_query.append(f'processinglevel:{processing_level}')
    if processing_mode:
        filter_query.append(f'processingmode:{processing_mode}')
    filter_query = ' AND '.join(filter_query)
    query = {'filter': filter_query, 'offset': offset, 'limit': limit,
             'sortedby': 'ingestiondate', 'order': 'desc'}
    query = urllib.parse.urlencode(query, safe='():,\\[]',
                                   quote_via=urllib.parse.quote)
    path = '/api/stub/products?' + query
    body = __http_request(path)
    return json.loads(body.decode('utf8'))


def search(polygon=None, begin_ts=None, end_ts=None, product=None,
           processing_level='L2', processing_mode=None, per_request_limit=25):
    '''Search for products via API.

    :param polygon: WKT polygon specifying an area the data should intersect
    :param begin_ts: Datetime specifying the earliest sensing date
    :param end_ts: Datetime specifying the latest sensing date
    :param product: Type of product to request
    :param processing_level: Data processing level (``L1B`` or ``L2``)
    :param processing_mode: Data processing mode (``Offline``,
                            ``Near real time`` or ``Reprocessing``)
    :param per_request_limit: Limit number of results per request
    :returns: Dictionary containing information about found products
    '''
    try:
        begin_ts = begin_ts.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    except AttributeError:
        pass
    try:
        end_ts = end_ts.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    except AttributeError:
        pass

    count = 0
    total = 1
    data = None
    logger.info('Searching for Sentinel-5 products')
    while count < total:
        s = _search(polygon, begin_ts, end_ts, product, processing_level,
                    processing_mode, count, per_request_limit)
        total = s.get('totalresults', 0)
        if data:
            data['products'].extend(s['products'])
            data['totalresults'] = total
        else:
            data = s
        count = len(data['products'])
        logger.debug('Received %s of %s data sets', count, total)
    logger.info('Found %s products', count)
    return data


def download(products, output_dir='.'):
    '''Download a set of products via API.

    :param products: List with product information (e.g. retrieved via search).
                     The list needs to contain dictionaries which must at least
                     have the fields `uuid` and `identifier`.
    :param output_dir: Directory to which the files will be downloaded.
    '''
    for product in products:
        uuid = product['uuid']
        filename = os.path.join(output_dir, product['identifier'] + '.nc')
        logger.info('Downloading %s to %s', uuid, filename)
        base_path = f"/odata/v1/Products('{uuid}')"

        # Check if file exist
        if os.path.exists(filename):
            # Skip download if checksum matches
            if __check_md5(filename, base_path):
                logger.info('Skipping %s since it already exist.', filename)
                continue
            logger.info('Overriding %s since md5 hash differs.', filename)

        # Download file
        __http_request(f'{base_path}/$value', filename)



# -*- coding: utf-8 -*-
'''
Sentinel-5p Downloader
~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2019, The Emissions API Developers
:url: https://emissions-api.org
:license: MIT
'''

import argparse
import dateutil.parser
import certifi
import logging
import multiprocessing
import textwrap
import sentinel5dl
from sentinel5dl import search, download

PRODUCTS = (
    'L1B_IR_SIR',
    'L1B_IR_UVN',
    'L1B_RA_BD1',
    'L1B_RA_BD2',
    'L1B_RA_BD3',
    'L1B_RA_BD4',
    'L1B_RA_BD5',
    'L1B_RA_BD6',
    'L1B_RA_BD7',
    'L1B_RA_BD8',
    'L2__AER_AI',
    'L2__AER_LH',
    'L2__CH4___',
    'L2__CLOUD_',
    'L2__CO____',
    'L2__HCHO__',
    'L2__NO2___',
    'L2__NP_BD3',
    'L2__NP_BD6',
    'L2__NP_BD7',
    'L2__O3_TCL',
    'L2__O3____',
    'L2__SO2___',
)

PRODUCTS_STR = textwrap.fill(', '.join(PRODUCTS),
                             subsequent_indent='  ',
                             initial_indent='  ')

PROCESSING_LEVELS = (
    'L1B',
    'L2'
)

PROCESSING_MODES = (
    'Offline',
    'Near real time',
    'Reprocessing'
)


def is_polygon(polygon):
    '''Validate if the supplied polygon string is in the necessary format to be
    used as part of a WKT polygon string.

    :param polygon: Polygon string in the form of lon1 lat1, lon2 lat2, ...
    :return: WKT polygon
    '''
    values = [value.strip() for value in polygon.split(',')]

    # Polygon must be at least a triangle
    if len(values) < 4:
        raise ValueError('Polygon must be at least a triangle')

    # Check if we got float pairs
    for value in values:
        if len([float(x) for x in value.split()]) != 2:
            raise ValueError('Polygon values must be pairs of numbers')

    # Check if we got a closed polygon
    if values[0] != values[-1]:
        raise ValueError('Polygon is not closed')

    return f'POLYGON(({polygon}))'


def main():
    # Configure logging in the library
    logging.basicConfig()
    logger = logging.getLogger(sentinel5dl.__name__)
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(
        description='Search for and download Sentinel-5P data files',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=f'AVAILABLE PRODUCTS\n{PRODUCTS_STR}'
    )

    # type= can use a callable, use that for most of this
    parser.add_argument(
        '--polygon',
        type=is_polygon,
        help='''Polygon defining an area by a set of coordinates.
            Example: 30.1 10.0, 40.0 40.1, 20 40, 10 20, 30.1 10.0'''
    )

    parser.add_argument(
        '--product',
        choices=PRODUCTS,
        metavar='PRODUCT',
        default='L2__CO____',
        help='Type of product to search for'
    )

    parser.add_argument(
        '--level',
        choices=PROCESSING_LEVELS,
        default='L2',
        help='Data processing level'
    )

    parser.add_argument(
        '--mode',
        choices=PROCESSING_MODES,
        help='Data processing mode'
    )

    parser.add_argument(
        '--begin-ts',
        default='2019-09-01T00:00:00.000Z',
        type=dateutil.parser.parse,
        help='''Timestamp specifying the earliest sensing date.
            Example: 2019-09-01T00:00:00.000Z'''
    )

    parser.add_argument(
        '--end-ts',
        default='2019-09-17T23:59:59.999Z',
        type=dateutil.parser.parse,
        help='''Timestamp specifying the latest sensing date.
            Example: 2019-09-17T23:59:59.999Z'''
    )

    parser.add_argument(
        '--use-certifi',
        action='store_true',
        help='''If a Certificate Authority (CA) bundle is not already supplied
            by your operating system, certifi provides an easy way of
            providing a cabundle.'''
    )

    parser.add_argument(
        '--worker',
        type=int,
        default=1,
        help='Number of parallel downloads',
    )

    parser.add_argument(
        'download_dir',
        metavar='download-dir',
        help='Download directory'
    )

    args = parser.parse_args()

    # Provide a Certificate Authority (CA) bundle
    if args.use_certifi:
        sentinel5dl.ca_info = certifi.where()

    # Search for Sentinel-5 products
    result = search(
        polygon=args.polygon,
        begin_ts=args.begin_ts,
        end_ts=args.end_ts,
        product=args.product,
        processing_level=args.level,
        processing_mode=args.mode
    )

    # Download found products to the download directory with number of workers
    with multiprocessing.Pool(args.worker) as p:
        p.starmap(download, map(
            lambda product: ((product,), args.download_dir),
            result.get('products')))


if __name__ == '__main__':
    main()



import os
import sentinel5dl
import tempfile
import unittest


testpath = os.path.dirname(os.path.abspath(__file__))


class TestSentinel5dl(unittest.TestCase):

    def _mock_http_request(self, path, filename=None):
        '''Mock HTTP requests to the ESA API
        '''
        # download
        if filename is not None:
            self._count_download += 1
            with open(filename, 'wb') as f:
                f.write(b'123')
            return

        # search request
        if path.startswith('/api/stub/products?'):
            self._count_search_request += 1
            with open(os.path.join(testpath, 'products.json'), 'rb') as f:
                return f.read()

        # checksum request
        if path.endswith('/Checksum/Value/$value'):
            self._count_checksum_request += 1
            # MD5 checksum for string `123`
            return b'202CB962AC59075B964B07152D234B70'

    def setUp(self):
        '''Patch cURL based operation in sentinel5dl so that we do not really
        make any HTTP requests and reset the request counters.
        '''
        setattr(sentinel5dl, '__http_request', self._mock_http_request)
        self._count_search_request = 0
        self._count_checksum_request = 0
        self._count_download = 0

    def test(self):
        '''Test search and download.
        '''
        result = sentinel5dl.search(
            polygon='POLYGON((7 49,13 49,13 52,7 52,7 49))',
            begin_ts='2019-09-01T00:00:00.000Z',
            end_ts='2019-09-17T23:59:59.999Z',
            product='L2__CO____')

        # The result returned by the mock contains four products but claims a
        # total of eight products, making sentinel5dl request resources twice.
        self.assertEqual(self._count_search_request, 2)
        self.assertEqual(result['totalresults'], 8)
        self.assertEqual(result['totalresults'], len(result['products']))

        products = result['products']
        with tempfile.TemporaryDirectory() as tmpdir:

            # prepare a file which is half-downloaded
            file_one = os.path.join(tmpdir, products[0]['identifier'] + '.nc')
            with open(file_one, 'wb') as f:
                f.write(b'12')

            sentinel5dl.download(products, tmpdir)

            # test files
            for product in products:
                filename = os.path.join(tmpdir, product['identifier'] + '.nc')
                with open(filename, 'rb') as f:
                    self.assertEqual(f.read(), b'123')

            # We should have downloaded four files and have an additional four
            # files storing md5 checksums
            self.assertEqual(len(os.listdir(tmpdir)), 8)

        # We should have four checksum requests. One for each file
        self.assertEqual(self._count_checksum_request, 4)
        # We should have downloaded four unique files
        self.assertEqual(self._count_download, 4)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
import os

path = os.path.abspath(os.path.dirname(__file__))


def read(filename):
    with open(os.path.join(path, filename), encoding='utf-8') as f:
        return f.read()


setup(
    name='sentinel5dl',
    version='1.0',
    description='Sentinel-5p Downloader',
    author='Emissions API Developers',
    license='MIT',
    url='https://github.com/emissions-api/sentinel5dl',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: GIS',
    ],
    install_requires=read('requirements.txt').split(),
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    entry_points={
        'console_scripts': [
            'sentinel5dl = sentinel5dl.__main__:main'
        ]
    }
)
